from typing import Iterator, List, Optional

from lambo.segmenter.lambo import Lambo

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.tokenizers.token import Token
from combo.data.tokenizers.tokenizer import Tokenizer

SPLIT_LEVELS = ["TURN", "SENTENCE", "NONE"]
DEFAULT_SPLIT_LEVEL = "SENTENCE"

IDX = 0


def _token_idx() -> Iterator[int]:
    global IDX
    IDX += 1
    yield IDX


def _reset_idx():
    global IDX
    IDX = 0


def _sentence_tokens(token: Token,
                     split_multiwords: Optional[bool] = None) -> List[Token]:
    if split_multiwords and len(token.subwords) > 0:
        subword_idxs = [next(_token_idx()) for _ in range(len(token.subwords))]
        multiword = (token.text, (subword_idxs[0], subword_idxs[-1]))
        tokens = [Token(idx=s_idx, text=subword, multiword=multiword) for (s_idx, subword)
                  in zip(subword_idxs, token.subwords)]
        return tokens
    else:
        return [Token(idx=next(_token_idx()), text=token.text, subwords=token.subwords)]


@Registry.register('lambo_tokenizer')
class LamboTokenizer(Tokenizer):
    @register_arguments
    def __init__(
            self,
            language: str = "English",
            default_split_level: str = DEFAULT_SPLIT_LEVEL,
            default_split_multiwords: bool = True
    ):
        self._language = language
        self.__tokenizer = Lambo.get(language)
        self.__default_split_level = default_split_level.upper()

        self.__default_split_multiwords = default_split_multiwords

    def tokenize(self,
                 text: str,
                 split_level: Optional[str] = None,
                 split_multiwords: Optional[bool] = None) -> List[List[Token]]:
        """
        Simple tokenization - ignoring the sentence splits
        :param text:
        :param split_level: split on turns, sentences, or no splitting (return one list of tokens)
        :param split_multiwords: split subwords into separate tokens (e.g. can't into ca, n't)
        :return:
        """
        _reset_idx()
        document = self.__tokenizer.segment(text)
        tokens = []

        split_level = split_level if split_level is not None else self.__default_split_level
        split_multiwords = split_multiwords if split_multiwords is not None else self.__default_split_multiwords

        if split_level.upper() == "TURN":
            for turn in document.turns:
                _reset_idx()
                sentence_tokens = []
                for sentence in turn.sentences:
                    for token in sentence.tokens:
                        sentence_tokens.extend(_sentence_tokens(token, split_multiwords))
                tokens.append(sentence_tokens)
        elif split_level.upper() == "SENTENCE":
            for turn in document.turns:
                for sentence in turn.sentences:
                    _reset_idx()
                    sentence_tokens = []
                    for token in sentence.tokens:
                        if len(token.subwords) > 0 and split_multiwords:
                            # @TODO this is a very dirty fix for Lambo model's shortcomings
                            # I noticed that for longer words with multiwords it tends to remove the last letter in the last multiword
                            # so this is a quick workaround to fix it

                            # check if subwords in token.subwords are consistent with token.text
                            if "".join(token.subwords) != token.text:
                                fixed_subwords = fix_subwords(token)
                                token.subwords = fixed_subwords
                        sentence_tokens.extend(_sentence_tokens(token, split_multiwords))
                    tokens.append(sentence_tokens)
        else:
            for turn in document.turns:
                _reset_idx()
                for sentence in turn.sentences:
                    for token in sentence.tokens:
                        tokens.extend(_sentence_tokens(token, split_multiwords))
            tokens = [tokens]

        return tokens

def fix_subwords(token: Token):
    fixed_subwords = []
    text_it = 0
    for i, subword in enumerate(token.subwords):
        if token.text[text_it:text_it + len(subword)] == subword:
            if i == len(token.subwords) - 1 and (text_it + len(subword) < len(token.text)):
                subword = token.text[text_it:]
            fixed_subwords.append(subword)
            text_it += len(subword)
        else:
            fixed_subwords.append(token.text[text_it:text_it + len(subword)])
            text_it += len(subword)
    return fixed_subwords