import pytorch_lightning as pl
from torch.utils.data import DataLoader
from .data.NerDataset import NerDataset
from pathlib import Path
from .data.NerTokenizer import NerTokenizer


class NerDataModule(pl.LightningDataModule):
    """
    NerDataModule is a PyTorch Lightning DataModule designed for Named Entity Recognition (NER) tasks.
    It handles data loading and preprocessing for training, validation, and testing datasets.

    Attributes:
    ----------
    path_data : Path
        The path to the directory containing NER dataset files.
    tokenizer : NerTokenizer
        An instance of the NerTokenizer class used for tokenizing the text data.
    batch_size : int, optional
        The batch size for data loading. Default is 32.
    encoding : str, optional
        The character encoding to use when reading dataset files. Default is 'utf-8'.
    num_workers : int, optional
        The number of CPU processes to use for data loading. Default is 1.

    """
    def __init__(self,
                 path_data: Path,
                 tokenizer: NerTokenizer,
                 batch_size: int = 32,
                 encoding: str = 'utf-8',
                 num_workers: int = 1):
        super().__init__()

        # Dataloader variables
        self.num_workers = num_workers
        self.batch_size = batch_size

        # Dataset variables
        self.path_data = path_data
        self.tokenizer = tokenizer
        self.encoding = encoding

        # Data
        self.train_dataset = None
        self.test_dataset = None
        self.dev_dataset = None

    def setup(self,
              stage=None) -> None:
        """
        Set up datasets for training, validation, and testing.
        """
        # Checking whether there is train data inside directory
        file_path = self.path_data / "train.txt"
        if file_path.is_file():
            self.train_dataset = NerDataset(tokenizer=self.tokenizer,
                                            file_path=file_path,
                                            encoding=self.encoding)

        # Checking whether there is dev data inside directory
        file_path = self.path_data / "dev.txt"
        if file_path.is_file():
            self.dev_dataset = NerDataset(tokenizer=self.tokenizer,
                                          file_path=file_path,
                                          encoding=self.encoding)

        # Checking whether there is test data inside directory
        file_path = self.path_data / "test.txt"
        if file_path.is_file():
            self.test_dataset = NerDataset(tokenizer=self.tokenizer,
                                           file_path=file_path,
                                           encoding=self.encoding)

    def train_dataloader(self) -> DataLoader:
        """
        Returns a PyTorch DataLoader for the training dataset.
        """
        return DataLoader(self.train_dataset,
                          batch_size=self.batch_size,
                          shuffle=True,
                          num_workers=self.num_workers)

    def val_dataloader(self) -> DataLoader:
        """
        Returns a PyTorch DataLoader for the validation  dataset.
        """
        return DataLoader(self.dev_dataset,
                          batch_size=self.batch_size,
                          shuffle=False,
                          num_workers=self.num_workers)

    def test_dataloader(self) -> DataLoader:
        """
        Returns a PyTorch DataLoader for the testing dataset.
        """
        return DataLoader(self.test_dataset,
                          batch_size=self.batch_size,
                          shuffle=False,
                          num_workers=self.num_workers)
