import pytorch_lightning as pl
from .metrics.MetricHolder import MetricHolder
import torch.nn as nn
import torch
from .data.SpecialTokens import SpecialTokens
from .embedders.BertEmbedder import BertEmbedder
from .embedders.ComboCharEmbedder import ComboCharEmbedder
from .classifiers.VanillaClassifier import VanillaClassifier
from .classifiers.CrfClassifier import CrfClassifier
from .utils.download import download_file
from .utils.constructors import construct_loss_from_config
from pathlib import Path
import json
from transformers import logging
# Fixing the issue with transformers BERT some weights being not used and causing warnings
logging.set_verbosity_error()

CHAR_PAD_TOKEN, CHAR_START_TOKEN, CHAR_END_TOKEN, CHAR_UNKNOWN_TOKEN, PAD_TOKEN, START_TOKEN, END_TOKEN = \
    SpecialTokens.CHAR_PAD_TOKEN, SpecialTokens.CHAR_START_TOKEN, SpecialTokens.CHAR_END_TOKEN, \
    SpecialTokens.CHAR_UNKNOWN_TOKEN, SpecialTokens.PAD_TOKEN, SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN


class NerModel(pl.LightningModule):
    def __init__(self,
                 loss_fn,
                 char_to_id_map: dict = None,
                 label_to_id_map: dict = None,
                 config: dict = None):
        super().__init__()

        ### Define architecture ###
        # Define Bert Embedder
        bert_embedder_params = config["model"]["bert_embedder"]
        self.bert_embedder = BertEmbedder(use_start_end_token=config["data"]["use_start_end_token"],
                                          **bert_embedder_params)

        # Define Character Embedder
        if config["data"]["use_char_level_embeddings"]:
            if config["model"]["char_embedder"]["type"] == "combo":
                char_embedding_dim = config["model"]["char_embedder"]["char_embedding_dim"]
                self.char_embedder = ComboCharEmbedder(vocab_size=len(char_to_id_map),
                                                       char_embedding_dim=char_embedding_dim,
                                                       padding_idx=char_to_id_map[CHAR_PAD_TOKEN])
        else:
            self.char_embedder = None

        # Define Classifier
        bert_output_dim = self.bert_embedder.output_dimension
        char_output_dim = self.char_embedder.output_dimension if self.char_embedder is not None else 0
        input_dim = bert_output_dim + char_output_dim
        classifier_params = config["model"]["classifier"]
        if config["model"]["classifier"]["type"] == "vanilla":
            self.classifier = VanillaClassifier(label_to_id_map=label_to_id_map,
                                                input_dim=input_dim,
                                                to_tag_space=classifier_params["to_tag_space"])
        elif config["model"]["classifier"]["type"] == "crf":
            self.classifier = CrfClassifier(label_to_id_map=label_to_id_map,
                                            input_dim=input_dim,
                                            to_tag_space=classifier_params["to_tag_space"])
        if "dropout" in config["model"]:
            self.dropout = nn.Dropout(p=config["model"]["dropout"])
        else:
            self.dropout = None
        ### End of architecture definition ###

        # Define metric holder
        self.train_metrics_holder = MetricHolder(label_to_idx=label_to_id_map)
        self.val_metrics_holder = MetricHolder(label_to_idx=label_to_id_map)
        self.test_metrics_holder = MetricHolder(label_to_idx=label_to_id_map)

        # Define optimization parameters
        self.learning_rate = config["learning_rate"]
        self.loss_fn = "crf" if config["model"]["classifier"]["type"] == "crf" else loss_fn

        # Save label to id map
        self.label_to_idx_map = label_to_id_map

        # save config
        self.config = config

    def forward(self, batch):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        word_ids = batch["word_ids"] if "word_ids" in batch else None

        # get BERT embeddings
        outputs = self.bert_embedder(input_ids=input_ids,
                                     attention_mask=attention_mask,
                                     word_ids=word_ids)
        # apply dropout
        if self.dropout:
            outputs = self.dropout(outputs)

        # adding character embeddings to the output of BERT
        if self.char_embedder:
            char_ids = batch["char_ids"]
            char_embeddings = self.char_embedder(char_ids)
            outputs = torch.cat((outputs, char_embeddings), dim=2)

        # Apply the classifier layer to obtain logits
        if self.loss_fn == "crf":
            _, logits = self.classifier(encoder_outputs=outputs)
        else:
            logits = self.classifier(outputs)

        return logits, outputs

    def training_step(self, batch, batch_idx):
        loss, preds, labels = self._common_step(batch)

        self.train_metrics_holder.update(predictions=preds,
                                         targets=labels)

        self.log(f"train_loss", loss, on_step=True, on_epoch=False, prog_bar=True, logger=True)
        return loss

    def on_train_epoch_end(self):
        # compute metrics with bio tags
        metrics_dict = self.train_metrics_holder.get_metrics(stage="train")
        metrics_dict["epoch"] = float(self.current_epoch + 1)

        # log metrics
        self.log_dict(metrics_dict, on_step=False, on_epoch=True, prog_bar=True, logger=True)

        # reset all metrics
        self.train_metrics_holder.reset()

    def validation_step(self, batch, batch_idx):
        loss, preds, labels = self._common_step(batch)

        self.val_metrics_holder.update(predictions=preds,
                                       targets=labels)

        self.log(f"validation_loss", loss, on_step=True, on_epoch=False, prog_bar=True, logger=True)
        return loss

    def on_validation_epoch_end(self):
        # compute metrics with bio tags
        metrics_dict = self.val_metrics_holder.get_metrics(stage="validation")
        metrics_dict["epoch"] = float(self.current_epoch + 1)

        # log metrics
        self.log_dict(metrics_dict, on_step=False, on_epoch=True, prog_bar=True, logger=True)

        # reset all metrics
        self.val_metrics_holder.reset()

    def test_step(self, batch, batch_idx):
        loss, preds, labels = self._common_step(batch)

        self.test_metrics_holder.update(predictions=preds,
                                        targets=labels)
        return loss

    def on_test_epoch_end(self):
        # compute metrics with bio tags
        metrics_dict = self.test_metrics_holder.get_metrics(stage="test")
        metrics_dict["epoch"] = float(self.current_epoch + 1)

        # log metrics
        self.log_dict(metrics_dict, on_step=False, on_epoch=True, prog_bar=True, logger=True)

        print(self.test_metrics_holder.get_classification_report())

        # reset all metrics
        self.test_metrics_holder.reset()

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(),
                                     lr=self.learning_rate)

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer,
                                                               mode="max",
                                                               patience=3,
                                                               factor=0.1,
                                                               threshold_mode="abs",
                                                               threshold=0.002)
        return {"optimizer": optimizer,
                "lr_scheduler": scheduler,
                "monitor": "validation_f1"}

    def _common_step(self, batch):
        labels = batch["labels_ids"]
        prediction_mask = batch["prediction_mask"]

        # Forward pass
        logits, encoder_outputs = self.forward(batch)

        # if CRF is used, we don't need to calculate the loss
        if self.loss_fn != "crf":
            # Get predictions
            preds = torch.argmax(logits, dim=2)
            # Permute the logits to conform loss requirements (batch_size, numb_classes, sequence_length)
            logits = logits.permute(0, 2, 1)
            # Change labels of PAD START and END tokens to -100
            labels = self.mask_labels(labels,prediction_mask)
            # Calculate the loss
            loss = self.loss_fn(logits, labels)
        else:
            loss = self.classifier.neg_log_likelihood(encoder_outputs, labels)
            # Change labels of PAD START and END tokens to -100
            labels = self.mask_labels(labels, prediction_mask)
            preds = logits

        return loss, preds, labels

    @staticmethod
    def mask_labels(labels,
                    prediction_mask):
        # for torch gradient computation we need to do it on copy
        label_copy = labels.clone()
        label_copy[prediction_mask == 0] = -100
        return label_copy

    def predict(self, batch):
        # Forward pass
        logits, _ = self.forward(batch)
        prediction_mask = batch["prediction_mask"]

        if self.loss_fn != "crf":
            # Get predictions
            preds = torch.argmax(logits, dim=2)
        else:
            preds = logits

        batch_size, sequence_length = preds.shape
        result = [[preds[i, j].item() for j in range(sequence_length) if prediction_mask[i, j] == 1] for i in range(batch_size)]
        return result

    @classmethod
    def get(cls, model_name):
        """
        Downloads and loads pretrained model from MinIO
        """
        # Downloading files and loading jsons
        PATH_CONFIG = Path(download_file(model_name, "config.json"))
        config = json.load(open(PATH_CONFIG))

        try:
            PATH_CHAR_TO_ID = Path(download_file(model_name, "char_to_id.json"))
            char_to_id = json.load(open(PATH_CHAR_TO_ID))
        except:
            char_to_id = {}

        PATH_LABEL_TO_ID = Path(download_file(model_name, "label_to_id.json"))
        label_to_id = json.load(open(PATH_LABEL_TO_ID))

        PATH_MODEL = Path(download_file(model_name, "best_model.ckpt"))
        # define model
        loss_fn = construct_loss_from_config(config=config,
                                             label_to_id=label_to_id)

        if torch.cuda.is_available():
            device = "cuda"
        else:
            device = "cpu"

        # load model
        model = NerModel.load_from_checkpoint(PATH_MODEL,
                                              loss_fn=loss_fn,
                                              char_to_id_map=char_to_id,
                                              map_location=torch.device(device),
                                              label_to_id_map=label_to_id,
                                              config=config)
        return model

    @classmethod
    def load_from_disc(cls,
                       folder_path: Path):
        """
        Loads pretrained model from disc
        """
        # Loading jsons
        PATH_CONFIG = folder_path / "config.json"
        config = json.load(open(PATH_CONFIG))

        try:
            PATH_CHAR_TO_ID = folder_path / "char_to_id.json"
            char_to_id = json.load(open(PATH_CHAR_TO_ID))
        except:
            char_to_id = {}

        PATH_LABEL_TO_ID = folder_path / "label_to_id.json"
        label_to_id = json.load(open(PATH_LABEL_TO_ID))

        PATH_MODEL = folder_path / "best_model.ckpt"
        # define model
        loss_fn = construct_loss_from_config(config=config,
                                             label_to_id=label_to_id)

        if torch.cuda.is_available():
            device = "cuda"
        else:
            device = "cpu"

        # load model
        model = NerModel.load_from_checkpoint(PATH_MODEL,
                                              loss_fn=loss_fn,
                                              char_to_id_map=char_to_id,
                                              map_location=torch.device(device),
                                              label_to_id_map=label_to_id,
                                              config=config)
        return model
