import torch
from torch import nn
import torch.nn.functional as F


class ComboCharEmbedder(nn.Module):
    """
    A character embedding model that utilizes convolutional layers to generate embeddings
    for sequences of characters in words. This model architecture for word character embeddings is
    the same as in COMBO (https://aclanthology.org/2021.emnlp-demo.7.pdf).

    Attributes:
    ----------
    embeddings : nn.Embedding
        The character embedding layer.
    conv1 : nn.Conv1d
        The first convolutional layer.
    conv2 : nn.Conv1d
        The second convolutional layer.
    conv3 : nn.Conv1d
        The third convolutional layer.
    output_dimension : int
        The dimension of the output embeddings.

    Methods:
    -------
    forward(batch_char_ids: torch.Tensor) -> torch.Tensor:
        Forward pass to generate embeddings for the given batch of character sequences.
    """
    def __init__(self,
                 vocab_size: int,
                 char_embedding_dim: int,
                 padding_idx: int = 0):
        """
        Initialize the ComboCharEmbedder model.

        Parameters:
        ----------
        vocab_size : int
            Size of the character vocabulary including special tokens.
        char_embedding_dim : int
            Dimension of the output embeddings.
        padding_idx : int, optional
            The index of the padding token in the vocabulary. Defaults to 0.
        """
        super(ComboCharEmbedder, self).__init__()

        # character embeddings
        self.embeddings = nn.Embedding(num_embeddings=vocab_size,
                                       embedding_dim=char_embedding_dim,
                                       padding_idx=padding_idx)

        # convolutional layers
        self.conv1 = nn.Conv1d(in_channels=char_embedding_dim,
                               out_channels=512,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               dilation=1)
        self.conv2 = nn.Conv1d(in_channels=512,
                               out_channels=256,
                               kernel_size=3,
                               stride=1,
                               padding=2,
                               dilation=2)
        self.conv3 = nn.Conv1d(in_channels=256,
                               out_channels=char_embedding_dim,
                               kernel_size=3,
                               stride=1,
                               padding=4,
                               dilation=4)

        # save output dimension
        self.output_dimension = char_embedding_dim

    def forward(self, batch_char_ids):
        """
        Perform a forward pass through the model. After convolutions max pooling across character embeddings is
        performed.

        Parameters:
        ----------
        batch_char_ids : torch.Tensor
            A tensor of character IDs for a batch of sequences (words). Input shape: batch_size x words x characters.

        Returns:
        -------
        torch.Tensor
            A tensor of embeddings for each sequence in the batch.
        """
        # embed characters
        x = self.embeddings(batch_char_ids)
        stacked_embeddings = []  # Empty list to store stacked word embeddings

        # Iterate over sentences
        for i in range(x.shape[0]):
            # sentence representation
            # transforms to words x embedding_dim x characters_length
            sentence = x[i].permute(0, 2, 1)

            # passing through convolutions
            sentence = F.relu(self.conv1(sentence))
            sentence = F.relu(self.conv2(sentence))
            sentence = self.conv3(sentence)

            # max pooling across dimensions for whole sequence
            # output: words_number x char_embedding_dim
            word_embeddings = torch.max(sentence, dim=2)[0]
            stacked_embeddings.append(word_embeddings)  # Append word embeddings to the list

        # combining results
        # output batch_size x words x char_embedding_dim
        stacked_embeddings = torch.stack(stacked_embeddings, dim=0)

        return stacked_embeddings
