from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from ..callbacks.FixedProgressBar import FixedProgressBar
from typing import Dict
import torch
from ..data.NerTokenizer import NerTokenizer
from ..NerDataModule import NerDataModule
from pathlib import Path
from ..loss.FocalLoss import focal_loss
from ..data.SpecialTokens import SpecialTokens

PAD_TOKEN, START_TOKEN, END_TOKEN = SpecialTokens.PAD_TOKEN, SpecialTokens.START_TOKEN, SpecialTokens.END_TOKEN


def construct_tokenizer_from_config(config: Dict,
                                    char_to_id_map: Dict,
                                    label_to_id_map: Dict) -> NerTokenizer:
    """
    Constructs a Named Entity Recognition tokenizer from a configuration dictionary.

    Parameters:
    -----------
    config : Dict
        Configuration dictionary containing all necessary parameters to initialize the tokenizer.
    char_to_id_map : Dict
        Mapping of characters to their respective unique identifiers.
    label_to_id_map : Dict
        Mapping of labels to their respective unique identifiers.

    Returns:
    --------
    NerTokenizer
        Initialized NER tokenizer object.
    """
    tokenizer = NerTokenizer(pretrained_model_type=config["model"]["bert_embedder"]["pretrained_model_type"],
                             pretrained_model_name=config["model"]["bert_embedder"]["pretrained_model_name"],
                             char_to_id_map=char_to_id_map,
                             label_to_id_map=label_to_id_map,
                             use_char_level_embeddings=config["data"]["use_char_level_embeddings"],
                             use_start_end_token=config["data"]["use_start_end_token"],
                             tokenize_entities=config["data"]["tokenize_entities"],
                             language=config["data"].get("language", "pl"),
                             max_word_len=config["data"].get("max_word_len", 69))
    return tokenizer


def construct_data_module_from_config(config: Dict,
                                      tokenizer: NerTokenizer) -> NerDataModule:
    """
    Constructs a Named Entity Recognition data module for PyTorch Lightning usage from a configuration dictionary.

    Parameters:
    -----------
    config : Dict
        Configuration dictionary containing all necessary parameters to initialize the data module.
    tokenizer : NerTokenizer
        NER tokenizer to be used for tokenization. It should be custom tokenizer from this repository.

    Returns:
    --------
    NerDataModule
        Initialized NER data module object.
    """
    datamodule = NerDataModule(path_data=Path(config["data"]["path_data"]),
                               tokenizer=tokenizer,
                               batch_size=config["data"]["batch_size"],
                               encoding=config["data"]["encoding"],
                               num_workers=config["data"]["num_workers"])
    return datamodule


def construct_loss_from_config(config: Dict,
                               label_to_id: Dict):
    """
     Constructs a loss function based on the given configuration dictionary.

     Parameters:
     -----------
     config : Dict
         Configuration dictionary containing the loss configuration.
     label_to_id : Dict
         Mapping of labels to their respective unique identifiers.

     Returns:
     --------
     loss
         Loss function, which could be "crf", CrossEntropyLoss, or a custom focal loss.

     """
    if config["model"]["classifier"]["type"] == "crf":
        loss = "crf"
    elif config["loss"] == "ce":
        loss = torch.nn.CrossEntropyLoss()
    elif config["loss"] == "focal":
        size = len(label_to_id) - len({PAD_TOKEN, START_TOKEN, END_TOKEN} & set(label_to_id.keys()))
        loss = focal_loss(size=size)
    else:
        raise ValueError(f"Loss {config['loss']} not supported")

    return loss


def construct_callbacks_from_config(config_callbacks: Dict) -> list:
    """
    Constructs a list of callbacks based on the given configuration dictionary. Supports following callbacks
    EarlyStopping, LearningRateMonitor, ModelCheckpoint, and FixedProgressBar.

    Parameters:
    -----------
    config_callbacks : Dict
        Configuration dictionary containing the callback configuration.

    Returns:
    --------
    list
        List of initialized callback objects.
    """
    callbacks = []
    for callback in config_callbacks:
        if callback == "EarlyStopping":
            callbacks.append(EarlyStopping(**config_callbacks[callback]))
        elif callback == "LearningRateMonitor":
            callbacks.append(LearningRateMonitor(**config_callbacks[callback]))
        elif callback == "ModelCheckpoint":
            pass
        elif callback == "FixedProgressBar":
            callbacks.append(FixedProgressBar())
        else:
            raise Exception(f"Callback {config_callbacks[callback]} not supported")
    # adding checkpoint callback at the end
    if "ModelCheckpoint" in config_callbacks:
        callbacks.append(ModelCheckpoint(**config_callbacks["ModelCheckpoint"]))

    return callbacks
