import errno
import logging
import os
import requests
import tqdm
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)

_URL = "http://s3.clarin-pl.eu/dspace/combo/ner/{model_name}/{file_name}"
_HOME_DIR = os.getenv("HOME", os.curdir)
_CACHE_DIR = os.path.join(_HOME_DIR, ".tmp_ner_model")


def download_file(model_name, file_name, force=False):
    _make_cache_dir(model_name)
    url = _URL.format(model_name=model_name, file_name=file_name)
    local_filename = url.split("/")[-1]
    local_filename = os.path.join(model_name, local_filename)
    location = os.path.join(_CACHE_DIR, local_filename)

    if os.path.exists(location) and not force:
        logger.debug("Using cached model.")
        return location

    chunk_size = 1024

    logger.info(f"Downloading {file_name} from {url}")

    try:
        with _requests_retry_session(retries=2).get(url, stream=True, timeout=600) as r:
            pbar = tqdm.tqdm(unit="B", total=int(r.headers.get("content-length")),
                             unit_divisor=chunk_size, unit_scale=True)
            with open(location, "wb") as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    f.write(chunk)
                    pbar.update(len(chunk))
            pbar.close()

    except requests.exceptions.RequestException as e:
        logger.error(f"Request error: {e}")
        raise
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        raise

    return location


def _make_cache_dir(model_name):
    cache_dir_model = os.path.join(_CACHE_DIR, model_name)
    os.makedirs(cache_dir_model, exist_ok=True)
    logger.info(f"Cache directory created at {cache_dir_model}")


def _requests_retry_session(
    retries=3,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 504),
):
    session = requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session