# Config file documentation

Ideally config file should contain as much information as possible so it is easier to track experiments. In this file description of keys is provided.
Config file format is a `.json` file with nested dictionaries. First level keys are:
- `data` - containes information about data processing
- `model` - nested dictionary with information about model architecture
- `loss` - it is a string that specifies loss function. Can be either `ce`, `focal` or `crf`
- `learning_rate` - it is a float that specifies learning rate used in training
- `callbacks` - containes names of callbacks used in training
- `trainer`

Defalut values are values which will be set if following code is run. 
```python
from combo.ner_modules.utils.utils import check_config_constraints
config = check_config_constraints(config)
```

---
`data`

It is required for this dictionary to contain path_data unless training is done using [train.py](..%2Ftrain.py) and data is passed as an argument.

Default values:

```json
  "data": {
    "path_data": "./data/pl/",
    "use_char_level_embeddings": true,
    "use_start_end_token": true,
    "tokenize_entities": true,
    "batch_size": 32,
    "encoding": "utf-8",
    "num_workers": 1
  }
```
- `use_start_end_token` [optional bool]- if true, start and end token will be added to each sentence
- `tokenize_entities` [optional bool] - if true, entities will be tokenized by backbone tokenizer. It is advised to set it to true.
- `encoding` [optional str] - encoding of data files

---
`model`

Specifies model architecture. It has up to 4 keys:
- `bert_embedder` [required dict]
- `char_embedder` [optional dict] - if not specified, no character level embeddings will be used
- `classifier` [required dict]
- `dropout` [optional float] - dropout probability applied to BERT embedder output

Example values for `bert_embedder`:
```json
"bert_embedder": {
    "projection_dimension": null,
    "freeze_bert": true,
    "token_pooling": true,
    "pooling_strategy": "max"
```
- `pretrained_model_name` [required] is a string of HuggingFace Bert model
- `pretrained_model_type` [required] is a string of HuggingFace Bert model type should either `AutoModel` or `BertModel`
- `projection_dimension` [optional int] if specified, output of BERT will be projected to this dimension
- `freeze_bert` [optional bool] if true, BERT weights will be frozen during training
- `pooling_strategy` [optional str] if `token_pooling` is true, this key specifies pooling strategy. Can be either `max` or `mean`

Example values for `char_embedder`:
```json
"char_embedder": {"type" : "combo",
                  "char_embedding_dim":  64
                 }
```
- `type` [required str] specifies type of character level embeddings. Currently, it supports only `combo`. But `contextual` is planned to be added.
- `char_embedding_dim` [optional int] specifies dimension of character level embeddings.
- `context_window` [optional int] if `type` is `contextual`, this key specifies context window size. To be added in future version.

Example values for `classifier`:
```json
"classifier": {"type" : "crf",
               "to_tag_space" :  "bilstm"},
```
- `type` [required str] specifies type of classifier. Can be either `crf` or `vanilla`
- `to_tag_space` [required str] specifies type of layer that transforms output of BERT and possibly character embedder to tag space. Can be either `bilstm`, `transformer` or `linear`

---
`callbacks`

Each key of callbacks dict is another callback used in training, despite `FixedProgressBar` which is a fix to common problem with progress bar in PyTorch Lightning, each callback is a callback implemented in PyTorch lightning library. Values of other thatn `FixedProgressBar` callbacks are dictionaries with arguments passed to callback. By default there is no callbacks used in training. 

Example values:
```json
"callbacks": {"FixedProgressBar": true,
            "LearningRateMonitor": {"logging_interval":"epoch"},
            "ModelCheckpoint": {"monitor": "validation_f1",
                                "mode": "max",
                                "save_top_k": 1,
                                "save_weights_only": true,
                                "filename" : "best_model"},
            "EarlyStopping": {"monitor": "validation_f1",
                              "mode": "max",
                              "min_delta": 0.001, 
                              "patience": 6}}
```
---
`trainer`

Similarly to `callbacks`, each key of `trainer` dict is a parameter of PyTorch Lightning trainer. 

Deafult values are:
```json
"trainer": {"devices": [0],
          "max_epochs": 50,
          "accelerator": "cuda",
          "log_every_n_steps": 10}
```